import logging
import os
import datetime
import torchvision.models as models
import math
import torch
import yaml
from easydict import EasyDict
import shutil
import glob
import pickle
from robustness import datasets as robust_dsets
from robustness import model_utils as robust_model_utils
import numpy as np
import matplotlib.pyplot as plt
import cv2

_ROBUST_MODEL_ROOT = ''#REDACTED
_IMAGENET_ROOT = ''#REDACTED

def cnt_params(model):
    cnt = 0
    for p in model.parameters():
        cnt += p.reshape(-1).shape[0]
    return cnt


def load_robust_resnet(mtype, just_the_model=True):
    resume_root = '{}/{}.ckpt'.format(_ROBUST_MODEL_ROOT, mtype)
    add_custom_forward = True
    if 'wide' in mtype:
        arch = models.wide_resnet50_2() if '50' in mtype else models.wide_resnet101_2()
    elif 'mobilenet' in mtype:
        arch = models.mobilenet_v2()
    elif 'shufflenet' in mtype:
        arch = models.shufflenet_v2_x1_0()
    elif 'vgg' in mtype:
        arch = models.vgg16_bn()
    elif 'densenet' in mtype:
        arch = models.densenet161()
    elif 'resnext50' in mtype:
        arch=models.resnext50_32x4d()
    else:
        _, arch, norm, suffix = mtype.split('_')
        eps = suffix[3:] # 'eps0.25'
        resume_root = '/{}_{}_eps{}.ckpt'.format(_ROBUST_MODEL_ROOT, arch, norm, eps)
        add_custom_forward = False
    # dataset_function = getattr(robustness.datasets, 'ImageNet')
    dataset_function = getattr(robust_dsets, 'ImageNet')
    dataset = dataset_function(_IMAGENET_ROOT)
    model_kwargs = {
            'arch': arch,
            'dataset': dataset,
            'resume_path': resume_root,
            'parallel': False,
            'add_custom_forward': add_custom_forward
        }

    model, _ = robust_model_utils.make_and_restore_model(**model_kwargs)
    model.eval()
    if just_the_model:
        model = model.model
    return model

def cache_results(results_path, results):
    with open(results_path, 'wb') as f:
        pickle.dump(results, f)

def load_cached_results(results_path):
    if os.path.exists(results_path):
        with open(results_path, 'rb') as f:
            results = pickle.load(f)
    else:
        results = dict()
    return results

def get_test_suite():
    timm_model_list = [
        'convit_tiny', 'convit_small', 'convit_base', 
        'swin_tiny_patch4_window7_224', 'swin_small_patch4_window7_224', 'swin_base_patch4_window7_224', 
        'deit_tiny_patch16_224', 'deit_small_patch16_224', 'deit_base_patch16_224',
        'mixer_b16_224', 
        'regnetx_002', 'regnetx_006', 'regnetx_008', 'regnetx_032', 'regnetx_064', 'regnetx_120', 'regnetx_320',
        'ssl_resnet18', 'ssl_resnet50', 'ssl_resnext50_32x4d', 'ssl_resnext101_32x4d', 'ssl_resnext101_32x8d', 'ssl_resnext101_32x16d',
        'resnext50_32x4d', 'resnext50d_32x4d', 'resnext101_32x8d',
        'resmlp_12_224', 'resmlp_24_224', 'resmlp_36_224', 'resmlp_12_distilled_224', 'resmlp_24_distilled_224', 'resmlp_36_distilled_224',
        'deit_base_distilled_patch16_224', 'deit_small_distilled_patch16_224', 'deit_tiny_distilled_patch16_224',
        'cait_xxs24_224', 'cait_xxs36_224', 'cait_s24_224', 
        'coat_lite_tiny', 'coat_lite_mini', 'coat_lite_small', 'coat_tiny', 'coat_mini', 

        # 'vit_tiny_patch16_224',
        'vit_small_patch16_224', 'vit_base_patch16_224',
        'efficientnet_b0', 'efficientnet_b1', 'efficientnet_b2', 'efficientnet_b3', 'efficientnet_b4',
        'resnet18', 'resnet34', 'resnet50', 'wide_resnet50_2', 'wide_resnet101_2', 'resnext50_32x4d',
        'vgg19', 'vgg19_bn', 'inception_v4', 'mnasnet_a1', 'mobilenetv2_100', 
    ]
    robust_models = [
        'robust_resnet18_l2_eps0.25', 'robust_resnet18_l2_eps1', 'robust_resnet18_l2_eps3',
        'robust_resnet50_l2_eps0.25', 'robust_resnet50_l2_eps1', 'robust_resnet50_l2_eps3',
        'robust_resnet18_linf_eps0.5', 'robust_resnet18_linf_eps2.0', 'robust_resnet18_linf_eps4.0', 
        'robust_resnet50_linf_eps0.5', 'robust_resnet50_linf_eps2.0', 'robust_resnet50_linf_eps4.0', 'pytorch_resnet50'
    ]
    clip_list = ['clip_ViT-B_16', 'clip_ViT-B_32', 'clip_RN50','clip_RN101']
    ft_clip = [x+'_finetuned' for x in clip_list]
    ss_models = ['moco_resnet50', 'moco_vit-s', 'moco_vit-b', 'dino_vits16', 'dino_vits8', 'dino_resnet50', 'simclr_resnet50x1', 'simclr_resnet50x2']
    return ['timm_'+x for x in timm_model_list] + robust_models + clip_list + ft_clip + ss_models

def mkey_to_category(mkey):
    mkey = mkey.lower()
    if 'vit' in mkey or 'swin' in mkey or 'deit' in mkey or 'coat' in mkey or 'cait' in mkey:
        if 'clip' in mkey:
            return 'CLIP ViTs' + (' Zero-Shot' if 'finetuned' not in mkey else ' Finetuned')
        elif 'moco' in mkey or 'dino' in mkey:
            return 'SS ViTs'
        else:
            return 'Sup ViTs'
    elif 'mixer' in mkey or 'resmlp' in mkey:
        return 'MLP-Based'
    elif 'robust' in mkey:
        return 'Robust ResNets'
        # if 'l2' in mkey:
        #     return '$\ell_2$ Robust ResNets'
        # else:
        #     return '$\ell_\infty$ Robust ResNets'
    elif 'simclr' in mkey or 'moco' in mkey or 'dino' in mkey or 'ssl' in mkey:
        return 'SS ResNets'
    elif 'clip' in mkey:
        return 'CLIP RNs' + (' Zero-Shot' if 'finetuned' not in mkey else ' Finetuned')
    elif 'resnet' in mkey or 'regnet' in mkey or 'resnext' in mkey:
        return 'Sup ResNets'
    else:
        return 'Other CNNs'

def get_category_to_mkeys_dict():
    mkeys = get_test_suite()
    mkeys_as_saved = get_test_suite_as_saved()
    # category_to_mkeys_dict = dict({c:[] for c in ['Sup ViTs', 'SS ViTs', 'CLIP ViTs', 'CLIP RNs', 'SS ResNets', 'ResNets', 'Other CNNs', 'Robust ResNets']})
    category_to_mkeys_dict = dict({c:[] for c in ['Sup ViTs', 'SS ViTs', 'CLIP ViTs Zero-Shot', 'CLIP ViTs Finetuned', 'CLIP RNs Zero-Shot', 'CLIP RNs Finetuned', 'SS ResNets', 'Sup ResNets', 'MLP-Based', 'Other CNNs', 'Robust ResNets']})
    # category_to_mkeys_dict = dict({c:[] for c in ['Sup ViTs', 'SS ViTs', 'CLIP ViTs Zero-Shot', 'CLIP ViTs Finetuned', 'CLIP RNs Zero-Shot', 'CLIP RNs Finetuned', 'SS ResNets', 'Sup ResNets', 'MLP-Based', 'Other CNNs', '$\ell_2$ Robust ResNets', '$\ell_\infty$ Robust ResNets']})
    for mkey, mkey_as_saved in zip(mkeys, mkeys_as_saved):
        category = mkey_to_category(mkey)
        if category not in category_to_mkeys_dict:
            category_to_mkeys_dict[category] = []
        category_to_mkeys_dict[category].append(mkey_as_saved)
    return category_to_mkeys_dict

def get_test_suite_as_saved():
    model_list = get_test_suite()
    for i, m in enumerate(model_list):
        if 'timm_' in m:
            model_list[i] = m[5:]
        elif 'clip_' in m:
            model_list[i] = m[5:].replace('_', '/')
        elif 'moco' in m:
            model_list[i] = m.replace('-s', '_small').replace('-b', '_base')
    return model_list

def nickname(mkey):
    mkey = mkey.title()
    mkey = mkey.replace('Timm_', '')
    mkey = mkey.replace('Resnet', 'ResNet')
    mkey = mkey.replace('Vit', 'ViT')
    mkey = mkey.replace('Ssl', 'SSL')
    mkey = mkey.replace('_', ' ')
    mkey = mkey.replace('Xxs', 'XXS')

    if 'distilled' in mkey:
        mkey = 'Distilled' + mkey[:mkey.index('distilled')-1]

    mkey = mkey.replace('Resmlp', 'ResMLP')

    # for size
    for size in ['Tiny', 'Small', 'Base']:
        if size in mkey:
            mkey = mkey[:mkey.index(size)] + f'({size[0].title()})'

    # for clip
    if '-B' in mkey or 'Rn' in mkey:
        mkey = 'CLIP ' + mkey
        if 'Finetuned' not in mkey:
            mkey = mkey + ' (Zero-Shot)'
        else:
            mkey = mkey[:(mkey.index('Finetuned')-1)] + ' (Finetuned)'
        mkey = mkey.replace('Rn', 'ResNet')

    # for robust models
    mkey = mkey.replace('L2', '$\ell_2$')
    mkey = mkey.replace('Linf', '$\ell_\infty$')
    mkey = mkey.replace('Eps', '$\epsilon=$')

    return mkey

def heatmap(accs, xlabels, ylabels, fig_save_root, title, fsr_is_dir=True, ax_labels=None):
    plt.clf()
    plt.figure(figsize=(3+6*len(xlabels),3+6*len(ylabels)), dpi=300)
    fig, ax = plt.subplots()

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)

    im = ax.imshow(accs, vmin=0, vmax=100)
    # We want to show all ticks...
    ax.set_xticks(np.arange(len(xlabels)))
    ax.set_yticks(np.arange(len(ylabels)))
    # ... and label them with the respective list entries
    ax.set_xticklabels(xlabels, fontsize=2.75)
    ax.set_yticklabels(ylabels, fontsize=2.75)
    b, t = plt.ylim() # discover the values for bottom and top
    b += 0.5 # Add 0.5 to the bottom
    t -= 0.5 # Subtract 0.5 from the top
    plt.ylim(b, t)
    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
            rotation_mode="anchor")
    thresh = 0#np.max(accs) * 0.5
    get_color = lambda x: "black" if x > thresh else "w"
    # Loop over data dimensions and create text annotations.
    for i in range(len(ylabels)):
        for j in range(len(xlabels)):
            # text = ax.text(j, i, np.round(accs[i, j],0), size=2,
            text = ax.text(j, i, int(accs[i, j]), size=1.75,
                        ha="center", va="center", color=get_color(accs[i,j]))
    if ax_labels is not None:
        xlabel, ylabel = ax_labels
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
    ax.set_title(title, fontsize=5)
    plt.tight_layout()
    fpath = '{}/heatmap.jpg' if fsr_is_dir else '{}_heatmap.jpg'
    plt.savefig(fpath.format(fig_save_root), bbox_inches='tight', pad_inches=0.001, dpi=500)
    plt.close()

### Utils for Generating and Visualizing NAMs

def compute_feature_maps(images, model, layer_name='layer4'):
    x = images.cuda()
    # normalizer_module = model._modules['normalizer']
    # feature_module = model._modules['model']
    # x = normalizer_module(images)
    # for name, module in model._modules.items():

    # print(model._modules.keys())
    for name, module in model.feat_net._modules.items():
        x = module(x)
        if name == layer_name:
            break
    return x

def compute_nams(model, images, feature_index, layer_name='layer4'):
    # print(images.shape)
    b_size = images.shape[0]
    feature_maps = compute_feature_maps(images, model, layer_name=layer_name)
    # print(feature_maps.shape)
    nams = (feature_maps[:, feature_index, :, :]).detach()
    nams_flat = nams.view(b_size, -1) 
    nams_max, _ = torch.max(nams_flat, dim=1, keepdim=True)
    nams_flat = nams_flat/nams_max
    nams = nams_flat.view_as(nams)
    nams_resized = []
    # print(nams.shape)
    # print(images.shape[2:])
    for nam in nams:
        nam = nam.cpu().numpy()
        # print(nam.shape)
        nam = cv2.resize(nam, images.shape[2:])
        nams_resized.append(nam)
    # print(nams[0].shape)
    nams = np.stack(nams_resized, axis=0)
    # print(nams.shape)
    nams = torch.from_numpy(1-nams)
    # print(nams.shape)
    return nams

def compute_heatmaps(imgs, masks):
    imgs = imgs.swapaxes(1,2).swapaxes(2,3)
    heatmaps = []
    # print(imgs.shape, masks.shape)
    for (img, mask) in zip(imgs, masks):
        heatmap = cv2.applyColorMap(np.uint8(255 * mask), cv2.COLORMAP_JET)
        heatmap = np.float32(heatmap) / 255
        heatmap = heatmap + np.float32(img)
        heatmap = heatmap / np.max(heatmap)
        heatmaps.append(heatmap)
    heatmaps = np.stack(heatmaps, axis=0)
    heatmaps = torch.from_numpy(heatmaps).permute(0, 3, 1, 2)
    return heatmaps


def grad_step(adv_inputs, grad, step_size):
    l = len(adv_inputs.shape) - 1
    grad_norm = torch.norm(grad.view(grad.shape[0], -1), dim=1).view(-1, *([1]*l))
    scaled_grad = grad / (grad_norm + 1e-10)
    return adv_inputs + scaled_grad * step_size


def feature_attack(model, seed_images, feature_indices, eps=500, 
                   step_size=10, iterations=100):
    seed_images = seed_images.cuda()
    batch_size = seed_images.shape[0]
    ftr_encoder = torch.nn.Sequential(*list(model.children())[:-1])
    for i in range(iterations+1):
        seed_images.requires_grad_()
        # (_, features), _ = model(seed_images, with_latent=True)
        features = ftr_encoder(seed_images).flatten(1)
        features_select = features[torch.arange(batch_size), feature_indices]
        if i==iterations:
            seed_images = seed_images.detach().cpu()
            features_select = features_select.detach().cpu().numpy()
            break   
        adv_loss = features_select.sum()
        grads = torch.autograd.grad(adv_loss, [seed_images])[0]
        seed_images = grad_step(seed_images.detach(), grads, step_size)
        seed_images = torch.clamp(seed_images, min=0., max=1.)
    return seed_images, features_select

imagenet_classes = ["tench", "goldfish", "great white shark", "tiger shark", "hammerhead shark", "electric ray", "stingray", "rooster", "hen", "ostrich", "brambling", "goldfinch", "house finch", "junco", "indigo bunting", "American robin", "bulbul", "jay", "magpie", "chickadee", "American dipper", "kite (bird of prey)", "bald eagle", "vulture", "great grey owl", "fire salamander", "smooth newt", "newt", "spotted salamander", "axolotl", "American bullfrog", "tree frog", "tailed frog", "loggerhead sea turtle", "leatherback sea turtle", "mud turtle", "terrapin", "box turtle", "banded gecko", "green iguana", "Carolina anole", "desert grassland whiptail lizard", "agama", "frilled-necked lizard", "alligator lizard", "Gila monster", "European green lizard", "chameleon", "Komodo dragon", "Nile crocodile", "American alligator", "triceratops", "worm snake", "ring-necked snake", "eastern hog-nosed snake", "smooth green snake", "kingsnake", "garter snake", "water snake", "vine snake", "night snake", "boa constrictor", "African rock python", "Indian cobra", "green mamba", "sea snake", "Saharan horned viper", "eastern diamondback rattlesnake", "sidewinder rattlesnake", "trilobite", "harvestman", "scorpion", "yellow garden spider", "barn spider", "European garden spider", "southern black widow", "tarantula", "wolf spider", "tick", "centipede", "black grouse", "ptarmigan", "ruffed grouse", "prairie grouse", "peafowl", "quail", "partridge", "african grey parrot", "macaw", "sulphur-crested cockatoo", "lorikeet", "coucal", "bee eater", "hornbill", "hummingbird", "jacamar", "toucan", "duck", "red-breasted merganser", "goose", "black swan", "tusker", "echidna", "platypus", "wallaby", "koala", "wombat", "jellyfish", "sea anemone", "brain coral", "flatworm", "nematode", "conch", "snail", "slug", "sea slug", "chiton", "chambered nautilus", "Dungeness crab", "rock crab", "fiddler crab", "red king crab", "American lobster", "spiny lobster", "crayfish", "hermit crab", "isopod", "white stork", "black stork", "spoonbill", "flamingo", "little blue heron", "great egret", "bittern bird", "crane bird", "limpkin", "common gallinule", "American coot", "bustard", "ruddy turnstone", "dunlin", "common redshank", "dowitcher", "oystercatcher", "pelican", "king penguin", "albatross", "grey whale", "killer whale", "dugong", "sea lion", "Chihuahua", "Japanese Chin", "Maltese", "Pekingese", "Shih Tzu", "King Charles Spaniel", "Papillon", "toy terrier", "Rhodesian Ridgeback", "Afghan Hound", "Basset Hound", "Beagle", "Bloodhound", "Bluetick Coonhound", "Black and Tan Coonhound", "Treeing Walker Coonhound", "English foxhound", "Redbone Coonhound", "borzoi", "Irish Wolfhound", "Italian Greyhound", "Whippet", "Ibizan Hound", "Norwegian Elkhound", "Otterhound", "Saluki", "Scottish Deerhound", "Weimaraner", "Staffordshire Bull Terrier", "American Staffordshire Terrier", "Bedlington Terrier", "Border Terrier", "Kerry Blue Terrier", "Irish Terrier", "Norfolk Terrier", "Norwich Terrier", "Yorkshire Terrier", "Wire Fox Terrier", "Lakeland Terrier", "Sealyham Terrier", "Airedale Terrier", "Cairn Terrier", "Australian Terrier", "Dandie Dinmont Terrier", "Boston Terrier", "Miniature Schnauzer", "Giant Schnauzer", "Standard Schnauzer", "Scottish Terrier", "Tibetan Terrier", "Australian Silky Terrier", "Soft-coated Wheaten Terrier", "West Highland White Terrier", "Lhasa Apso", "Flat-Coated Retriever", "Curly-coated Retriever", "Golden Retriever", "Labrador Retriever", "Chesapeake Bay Retriever", "German Shorthaired Pointer", "Vizsla", "English Setter", "Irish Setter", "Gordon Setter", "Brittany dog", "Clumber Spaniel", "English Springer Spaniel", "Welsh Springer Spaniel", "Cocker Spaniel", "Sussex Spaniel", "Irish Water Spaniel", "Kuvasz", "Schipperke", "Groenendael dog", "Malinois", "Briard", "Australian Kelpie", "Komondor", "Old English Sheepdog", "Shetland Sheepdog", "collie", "Border Collie", "Bouvier des Flandres dog", "Rottweiler", "German Shepherd Dog", "Dobermann", "Miniature Pinscher", "Greater Swiss Mountain Dog", "Bernese Mountain Dog", "Appenzeller Sennenhund", "Entlebucher Sennenhund", "Boxer", "Bullmastiff", "Tibetan Mastiff", "French Bulldog", "Great Dane", "St. Bernard", "husky", "Alaskan Malamute", "Siberian Husky", "Dalmatian", "Affenpinscher", "Basenji", "pug", "Leonberger", "Newfoundland dog", "Great Pyrenees dog", "Samoyed", "Pomeranian", "Chow Chow", "Keeshond", "brussels griffon", "Pembroke Welsh Corgi", "Cardigan Welsh Corgi", "Toy Poodle", "Miniature Poodle", "Standard Poodle", "Mexican hairless dog (xoloitzcuintli)", "grey wolf", "Alaskan tundra wolf", "red wolf or maned wolf", "coyote", "dingo", "dhole", "African wild dog", "hyena", "red fox", "kit fox", "Arctic fox", "grey fox", "tabby cat", "tiger cat", "Persian cat", "Siamese cat", "Egyptian Mau", "cougar", "lynx", "leopard", "snow leopard", "jaguar", "lion", "tiger", "cheetah", "brown bear", "American black bear", "polar bear", "sloth bear", "mongoose", "meerkat", "tiger beetle", "ladybug", "ground beetle", "longhorn beetle", "leaf beetle", "dung beetle", "rhinoceros beetle", "weevil", "fly", "bee", "ant", "grasshopper", "cricket insect", "stick insect", "cockroach", "praying mantis", "cicada", "leafhopper", "lacewing", "dragonfly", "damselfly", "red admiral butterfly", "ringlet butterfly", "monarch butterfly", "small white butterfly", "sulphur butterfly", "gossamer-winged butterfly", "starfish", "sea urchin", "sea cucumber", "cottontail rabbit", "hare", "Angora rabbit", "hamster", "porcupine", "fox squirrel", "marmot", "beaver", "guinea pig", "common sorrel horse", "zebra", "pig", "wild boar", "warthog", "hippopotamus", "ox", "water buffalo", "bison", "ram (adult male sheep)", "bighorn sheep", "Alpine ibex", "hartebeest", "impala (antelope)", "gazelle", "arabian camel", "llama", "weasel", "mink", "European polecat", "black-footed ferret", "otter", "skunk", "badger", "armadillo", "three-toed sloth", "orangutan", "gorilla", "chimpanzee", "gibbon", "siamang", "guenon", "patas monkey", "baboon", "macaque", "langur", "black-and-white colobus", "proboscis monkey", "marmoset", "white-headed capuchin", "howler monkey", "titi monkey", "Geoffroy's spider monkey", "common squirrel monkey", "ring-tailed lemur", "indri", "Asian elephant", "African bush elephant", "red panda", "giant panda", "snoek fish", "eel", "silver salmon", "rock beauty fish", "clownfish", "sturgeon", "gar fish", "lionfish", "pufferfish", "abacus", "abaya", "academic gown", "accordion", "acoustic guitar", "aircraft carrier", "airliner", "airship", "altar", "ambulance", "amphibious vehicle", "analog clock", "apiary", "apron", "trash can", "assault rifle", "backpack", "bakery", "balance beam", "balloon", "ballpoint pen", "Band-Aid", "banjo", "baluster / handrail", "barbell", "barber chair", "barbershop", "barn", "barometer", "barrel", "wheelbarrow", "baseball", "basketball", "bassinet", "bassoon", "swimming cap", "bath towel", "bathtub", "station wagon", "lighthouse", "beaker", "military hat (bearskin or shako)", "beer bottle", "beer glass", "bell tower", "baby bib", "tandem bicycle", "bikini", "ring binder", "binoculars", "birdhouse", "boathouse", "bobsleigh", "bolo tie", "poke bonnet", "bookcase", "bookstore", "bottle cap", "hunting bow", "bow tie", "brass memorial plaque", "bra", "breakwater", "breastplate", "broom", "bucket", "buckle", "bulletproof vest", "high-speed train", "butcher shop", "taxicab", "cauldron", "candle", "cannon", "canoe", "can opener", "cardigan", "car mirror", "carousel", "tool kit", "cardboard box / carton", "car wheel", "automated teller machine", "cassette", "cassette player", "castle", "catamaran", "CD player", "cello", "mobile phone", "chain", "chain-link fence", "chain mail", "chainsaw", "storage chest", "chiffonier", "bell or wind chime", "china cabinet", "Christmas stocking", "church", "movie theater", "cleaver", "cliff dwelling", "cloak", "clogs", "cocktail shaker", "coffee mug", "coffeemaker", "spiral or coil", "combination lock", "computer keyboard", "candy store", "container ship", "convertible", "corkscrew", "cornet", "cowboy boot", "cowboy hat", "cradle", "construction crane", "crash helmet", "crate", "infant bed", "Crock Pot", "croquet ball", "crutch", "cuirass", "dam", "desk", "desktop computer", "rotary dial telephone", "diaper", "digital clock", "digital watch", "dining table", "dishcloth", "dishwasher", "disc brake", "dock", "dog sled", "dome", "doormat", "drilling rig", "drum", "drumstick", "dumbbell", "Dutch oven", "electric fan", "electric guitar", "electric locomotive", "entertainment center", "envelope", "espresso machine", "face powder", "feather boa", "filing cabinet", "fireboat", "fire truck", "fire screen", "flagpole", "flute", "folding chair", "football helmet", "forklift", "fountain", "fountain pen", "four-poster bed", "freight car", "French horn", "frying pan", "fur coat", "garbage truck", "gas mask or respirator", "gas pump", "goblet", "go-kart", "golf ball", "golf cart", "gondola", "gong", "gown", "grand piano", "greenhouse", "radiator grille", "grocery store", "guillotine", "hair clip", "hair spray", "half-track", "hammer", "hamper", "hair dryer", "hand-held computer", "handkerchief", "hard disk drive", "harmonica", "harp", "combine harvester", "hatchet", "holster", "home theater", "honeycomb", "hook", "hoop skirt", "gymnastic horizontal bar", "horse-drawn vehicle", "hourglass", "iPod", "clothes iron", "carved pumpkin", "jeans", "jeep", "T-shirt", "jigsaw puzzle", "rickshaw", "joystick", "kimono", "knee pad", "knot", "lab coat", "ladle", "lampshade", "laptop computer", "lawn mower", "lens cap", "letter opener", "library", "lifeboat", "lighter", "limousine", "ocean liner", "lipstick", "slip-on shoe", "lotion", "music speaker", "loupe magnifying glass", "sawmill", "magnetic compass", "messenger bag", "mailbox", "tights", "one-piece bathing suit", "manhole cover", "maraca", "marimba", "mask", "matchstick", "maypole", "maze", "measuring cup", "medicine cabinet", "megalith", "microphone", "microwave oven", "military uniform", "milk can", "minibus", "miniskirt", "minivan", "missile", "mitten", "mixing bowl", "mobile home", "ford model t", "modem", "monastery", "monitor", "moped", "mortar and pestle", "graduation cap", "mosque", "mosquito net", "vespa", "mountain bike", "tent", "computer mouse", "mousetrap", "moving van", "muzzle", "metal nail", "neck brace", "necklace", "baby pacifier", "notebook computer", "obelisk", "oboe", "ocarina", "odometer", "oil filter", "pipe organ", "oscilloscope", "overskirt", "bullock cart", "oxygen mask", "product packet / packaging", "paddle", "paddle wheel", "padlock", "paintbrush", "pajamas", "palace", "pan flute", "paper towel", "parachute", "parallel bars", "park bench", "parking meter", "railroad car", "patio", "payphone", "pedestal", "pencil case", "pencil sharpener", "perfume", "Petri dish", "photocopier", "plectrum", "Pickelhaube", "picket fence", "pickup truck", "pier", "piggy bank", "pill bottle", "pillow", "ping-pong ball", "pinwheel", "pirate ship", "drink pitcher", "block plane", "planetarium", "plastic bag", "plate rack", "farm plow", "plunger", "Polaroid camera", "pole", "police van", "poncho", "pool table", "soda bottle", "plant pot", "potter's wheel", "power drill", "prayer rug", "printer", "prison", "missile", "projector", "hockey puck", "punching bag", "purse", "quill", "quilt", "race car", "racket", "radiator", "radio", "radio telescope", "rain barrel", "recreational vehicle", "fishing casting reel", "reflex camera", "refrigerator", "remote control", "restaurant", "revolver", "rifle", "rocking chair", "rotisserie", "eraser", "rugby ball", "ruler measuring stick", "sneaker", "safe", "safety pin", "salt shaker", "sandal", "sarong", "saxophone", "scabbard", "weighing scale", "school bus", "schooner", "scoreboard", "CRT monitor", "screw", "screwdriver", "seat belt", "sewing machine", "shield", "shoe store", "shoji screen / room divider", "shopping basket", "shopping cart", "shovel", "shower cap", "shower curtain", "ski", "balaclava ski mask", "sleeping bag", "slide rule", "sliding door", "slot machine", "snorkel", "snowmobile", "snowplow", "soap dispenser", "soccer ball", "sock", "solar thermal collector", "sombrero", "soup bowl", "keyboard space bar", "space heater", "space shuttle", "spatula", "motorboat", "spider web", "spindle", "sports car", "spotlight", "stage", "steam locomotive", "through arch bridge", "steel drum", "stethoscope", "scarf", "stone wall", "stopwatch", "stove", "strainer", "tram", "stretcher", "couch", "stupa", "submarine", "suit", "sundial", "sunglasses", "sunglasses", "sunscreen", "suspension bridge", "mop", "sweatshirt", "swim trunks / shorts", "swing", "electrical switch", "syringe", "table lamp", "tank", "tape player", "teapot", "teddy bear", "television", "tennis ball", "thatched roof", "front curtain", "thimble", "threshing machine", "throne", "tile roof", "toaster", "tobacco shop", "toilet seat", "torch", "totem pole", "tow truck", "toy store", "tractor", "semi-trailer truck", "tray", "trench coat", "tricycle", "trimaran", "tripod", "triumphal arch", "trolleybus", "trombone", "hot tub", "turnstile", "typewriter keyboard", "umbrella", "unicycle", "upright piano", "vacuum cleaner", "vase", "vaulted or arched ceiling", "velvet fabric", "vending machine", "vestment", "viaduct", "violin", "volleyball", "waffle iron", "wall clock", "wallet", "wardrobe", "military aircraft", "sink", "washing machine", "water bottle", "water jug", "water tower", "whiskey jug", "whistle", "hair wig", "window screen", "window shade", "Windsor tie", "wine bottle", "airplane wing", "wok", "wooden spoon", "wool", "split-rail fence", "shipwreck", "sailboat", "yurt", "website", "comic book", "crossword", "traffic or street sign", "traffic light", "dust jacket", "menu", "plate", "guacamole", "consomme", "hot pot", "trifle", "ice cream", "popsicle", "baguette", "bagel", "pretzel", "cheeseburger", "hot dog", "mashed potatoes", "cabbage", "broccoli", "cauliflower", "zucchini", "spaghetti squash", "acorn squash", "butternut squash", "cucumber", "artichoke", "bell pepper", "cardoon", "mushroom", "Granny Smith apple", "strawberry", "orange", "lemon", "fig", "pineapple", "banana", "jackfruit", "cherimoya (custard apple)", "pomegranate", "hay", "carbonara", "chocolate syrup", "dough", "meatloaf", "pizza", "pot pie", "burrito", "red wine", "espresso", "tea cup", "eggnog", "mountain", "bubble", "cliff", "coral reef", "geyser", "lakeshore", "promontory", "sandbar", "beach", "valley", "volcano", "baseball player", "bridegroom", "scuba diver", "rapeseed", "daisy", "yellow lady's slipper", "corn", "acorn", "rose hip", "horse chestnut seed", "coral fungus", "agaric", "gyromitra", "stinkhorn mushroom", "earth star fungus", "hen of the woods mushroom", "bolete", "corn cob", "toilet paper"]
